//
//   DVector class definition.  Copy on write version.
//
#ifndef DVECTYPE_HH
#define DVECTYPE_HH

#include "DVector.hh"
#include "CWVec.hh"

#ifdef __GNUG__
#pragma interface
#endif

/**  The typed data vector template describes the classes based on 
  *        DVector which contain vectors of variously typed data.
  *  @brief Typed data vector class template.
  */
template<class T>
class DVecType : public DVector {
public:
   using DVector::Extract; 
   
  /**  Data type used internally for size and index variables.
    *  @brief Size and index data type.
    */
  typedef DVector::size_type size_type;
  
  /**  Empty vector constructor.
    */
  DVecType(void);

  /**  Create a DVector with the same length and data as the 
    *        argument.
    *  @brief Copy constructor.
    */
  explicit DVecType(const DVector& dv);

  /**  Create a DVector with the same data as the argument, but a different
    *  length. This constructor
    *  @brief Copy constructor.
    */
  DVecType(const DVecType& dv, size_type len);

  /**  A data vector is constructed with the specified size. If an 
    *  initialization array is specified, the data are copied to the 
    *  vector. The current length is set to 'len'.
    *  @brief Data constructor.
    *  @param len  Vector length in data words.
    *  @param data Optional pointer to initial data.
    */
  explicit DVecType(size_type len, const T* data=0);

  /**  A data vector  is constructed from a specified substring of a 
    *  CWVec.
    *  @param cv containing data to be extracted.
    *  @param i  start position of data.
    *  @param l  length of data to be inserted.
    */
  explicit DVecType(const CWVec<T>& cv, size_type i=0, size_type l=0);

  /**  Delete a typed DVector.
    *  @brief Destructor.
    */
  ~DVecType();

  /**  Allocate or expand the vector to make room for 'len' words.
    *  Existing data are preserved, but no post padding is performed. 
    *  The current data length is set to 'len'. No write access is 
    *  provided to the data. Use <tt>Extend()</tt> to pad the vector to the 
    *  specified length. 
    *  @brief  Resize the data vector.
    *  @param len Minimum data vector size.
    */
  void ReSize(size_type len);

  /**  Insure that at least the specified number of data words are available.
    *  @brief  Reserve the specified number of data words.
    *  @param len Number of valid data words.
    */
  void reserve(size_type len);
 
  /**  Return the capacity of the data vector in data words.
    *  @brief Get the vector capacity
    *  @return Number of words that can fit in the vector.
    */
  size_type capacity(void) const;

  /**  Test whether the underlying copy-on-write data vector is shared with
    *  another DVecType object.
    *  @brief Test for shared data vector.
    *  @return true if data vector is shared.
    */
  bool shared(void) const;

  /**  Replace each element with its complex conjugate.
    *  @brief  Complex conjugate.
    */
  void Conjugate(void); 

  /**  A substring is removed from the vector.
    *  @brief Erase a substring from a vector.
    */
  void Erase(size_type inx, size_type len);

  /**  Extend the data vector to the specified length. If the current length 
    *  is less than the requested length, no action is taken. If the requested
    *  length is greater than the current length the vector is padded out 
    *  to the requested length with zeroes.
    *  @brief Extend a vector with zero padding.
    *  @param len new length.
    */
  void Extend(size_type len);

  /**  A new vector is generated from a substring of the vector data.
    *  @brief Extract a substring from a vector.
    */
  DVecType* Extract(size_type inx, size_type len) const;

  /**  A new vector is generated from a decimated substring of the vector 
    *        data.
    *  @brief Extract and decimate a substring from a vector.
    */
  DVecType* Extract(size_type inx, size_type len, size_type inc) const;

  /**  Return true if all elements of the vector are finite, i.e. not inf
    *  or NaN.
    *  @brief Test for finite data.
    *  @return true if all elements are finite.
    */
  bool finite(void) const;

  /**  Return true if all elements of the vector are normalized, i.e. not
    *  subnormal, inf or NaN.
    *  @brief Test for normalized data.
    *  @return true if all elements are normalized.
    */
  bool normal(void) const;

  /**  Returns the number of valid data words in the Data Vector.
    *  @brief Get the data length.
    *  @return The number of valid data words.
    */
  size_type getLength(void) const;

  /**  getSize return the size of a single data word in bytes.
    *  @brief Get data word size.
    */
  DVector::size_type getSize(void) const {return sizeof(T);}

  /**  Returns the data type as represented by the DVect::DVType 
    *        enumerator.
    *  @brief Get the data type.
    */
  inline DVType getType(void) const;
  static DVType getDataType(void);

  /**  Returns a (void*) pointer to the internal array. Both const and
    *        non-const references are available.
    *  @brief Get a data pointer.
    */
  void*       refData(void)       {return (void*) refTData();}
  const void* refData(void) const {return (const void*) refTData();}

  /**  Returns a writeable pointer to the internal array.
    *  @brief Get a writeable data pointer.
    */
  T*       refTData(void);

  /**  Returns a constant pointer to the internal array.
    *  @brief Get a constant data pointer.
    */
  const T* refTData(void) const;

  /**  The specified words of the data vector are copied to an
    *        array ('data') and the number of words copied is returned. 
    *        The data will be converted to the type of 'data' differs 
    *        from that of the vector. The second argument may be a short, 
    *        float or fComplex array.
    *  @brief Copy data vector to an array.
    */
  size_type getData(size_type inx, size_type len, short data[]) const;
  size_type getData(size_type inx, size_type len, int data[]) const;
  size_type getData(size_type inx, size_type len, float data[]) const;
  size_type getData(size_type inx, size_type len, double data[]) const;
  size_type getData(size_type inx, size_type len, fComplex data[]) const;
  size_type getData(size_type inx, size_type len, dComplex data[]) const;

  /**  A data word is picked up from the DVector and converted to the 
    *  named type.
    *  @brief Get one data word.
    */
  short getShort(size_type inx) const;

  /**  A data word is picked up from the DVector and converted to the 
    *  named type.
    *  @brief Get one data word.
    */
  int getInt(size_type inx) const;

  /**  A data word is picked up from the DVector and converted to the 
    *  named type.
    *  @brief Get one data word.
    */
  float getFloat(size_type inx) const;

  /**  A data word is picked up from the DVector and converted to the 
    *  named type.
    *  @brief Get one data word.
    */
  double getDouble(size_type inx) const;

  /**  A data word is picked up from the DVector and converted to the 
    *  named type.
    *  @brief Get one data word.
    */
  fComplex getCplx(size_type inx) const;

  /**  A data word is picked up from the DVector and converted to the 
    *  named type.
    *  @brief Get one data word.
    */
  dComplex getDCplx(size_type inx) const;

  /**  Return the largest entry. If the vector is complex, the largest 
    *  real part is returned.
    *  @brief   Maximum data value.
    *  @return Largest data value.
    */
  double getMaximum(void) const;

  /**  Return the smallest entry. If the vector is complex, the smallest 
    *  real part is returned.
    *  @brief   Minimum data value.
    *  @return Smallest data value.
    */
  double getMinimum(void) const;

  /**  Count the number of entries with value greater than the specified 
    *          limit. If the vector is complex, the real part is compared.
    *  @brief   Number of entries in range.
    *  @param  low  Lower limit of entries to be counted.
    *  @param  high Upper limit of entries to be counted.
    *  @return Number of entries greater than limit.
    */
  size_type getNBetween(double low, double high) const;

  /**  Count the number of entries with value greater than the specified 
    *          limit. If the vector is complex, the real part is compared.
    *  @brief   Number of entries greater than limit.
    *  @param  limit Lower limit of entries to be counted.
    *  @return Number of entries greater than limit.
    */
  size_type getNGreater(double limit) const;

  /**  Count the number of entries with value less than the specified 
    *          limit. If the vector is complex, the real part is compared.
    *  @brief   Number of entries less than limit.
    *  @param  limit Upper limit of entries to be counted.
    *  @return Number of entries less than limit.
    */
  size_type getNLess(double limit) const;

  /**  Replace <tt>N</tt> values starting at <tt>inx</tt> of the current 
    *  vector with <tt>Nv</tt> values from the argument vector starting at 
    *  index <tt>inv</tt>.
    *  @brief Inset data from a vector into the curret vector.
    *  @param inx Index of the insertion point of the current vector.
    *  @param N   Number of words to be inserted
    *  @param v   Vector containing data to be inserted.
    *  @param inv index in v of the data to be inserted.
    *  @param Nv  Number of words to be inserted
    *  @return Reference to the result vector.
    */
  DVecType& replace(size_type inx, size_type N, const DVector& v, 
		    size_type inv, size_type Nv);

  /**  Replace <tt>N</tt> values starting at <tt>inx</tt> of the current 
    *  vector with the first <tt>Nv</tt> words of vector <tt>v</tt>.
    *  @brief Inset data from a vector into the curret vector.
    *  @param inx Index of the insertion point of the current vector.
    *  @param N   Number of words to be replaced
    *  @param v   Data vector to be inserted.
    *  @param Nv  Number of words to be inserted
    *  @return Reference to the result vector.
    */
  DVecType& replace(size_type inx, size_type N, T v, size_type Nv);

  /**  Replace/insert <tt>N</tt> values starting at <tt>inx</tt> of the current vector with
    *  the specified data in reverse order.
    *  @brief Inset data from a vector into the curret vector.
    *  @param inx Index of the insertion point of the current vector.
    *  @param v   Data to be inserted in reverse order.
    *  @param N   Number of words to be inserted
    *  @return Reference to the result vector.
    */
  DVecType& reverse(size_type inx, const T* v, size_type N);

  /**  Add <tt>N</tt> values in the addend vector to the current vector 
    *  starting at index <tt>inx</tt> of the current vector and 
    *  <tt>inx2</tt> of the addend vector.
    *  @brief Add two vectors
    *  @param inx  Index of the first element in this vector to be added.
    *  @param v    Vector to be added
    *  @param inx2 index in the 
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  DVecType& add(size_type inx, const DVector& v, size_type inx2=0, 
		size_type N=0);

  /**  Add a constant to <tt>N</tt> values of the current vector starting at
    *  index <tt>inx</tt>.
    *  @brief Add a constant to a vector
    *  @param inx  Index of the first element in this vector to be added.
    *  @param x    Constant to be added
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  DVecType& bias(size_type inx, double x, size_type N=0);

  /**  Subtract <tt>N</tt> values in the subtrahend vector from the current 
    *  vector starting at index <tt>inx</tt> of the current vector and 
    *  <tt>inx2</tt> of the subtrahend vector.
    *  @brief Subtract two vectors
    *  @param inx  Index of the first element in this vector to be subtracted.
    *  @param v    Vector to be subtracted
    *  @param inx2 index in the subtrahend vector
    *  @param N    Number of words to be subtracted
    *  @return Reference to the result vector.
    */
  DVecType& sub(size_type inx, const DVector& v, size_type inx2=0, 
		size_type N=0);

  /**  Multiply <tt>N</tt> values starting at index <tt>inx</tt> of the 
    *  current vector by the values in the multiplier vector starting at 
    *  index <tt>inx2</tt>.
    *  @brief Multiply two vectors
    *  @param inx  Index of the first element in this vector to be multiplied.
    *  @param v    Vector to be multiplied
    *  @param inx2 index of first word in the current vector
    *  @param N    Number of words to be multiplied.
    *  @return Reference to the result vector.
    */
  DVecType& mpy(size_type inx, const DVector& v, size_type inx2=0, 
		size_type N=0);

  /**  Multiply <tt>N</tt> values starting at index <tt>inx</tt> of the 
    *  current vector by the cc values in the multiplier vector starting 
    *  at index <tt>inx2</tt>.
    *  @brief Multiply a vector by the complex conjugate of another.
    *  @param inx  Index of the first element in this vector to be multiplied.
    *  @param v    Vector to be multiplied
    *  @param inx2 index of first word in the current vector
    *  @param N    Number of words to be multiplied.
    *  @return Reference to the result vector.
    */
  DVecType& cmpy(size_type inx, const DVector& v, size_type inx2=0, 
		size_type N=0);

  /**  Multiply <tt>N</tt> values of the current vector starting at 
    *  index <tt>inx</tt> by a constant.
    *  @brief Multiply a vector by a constant
    *  @param inx  Index of the first element in this vector to be added.
    *  @param x    Constant to be added
    *  @param N    Number of words to be added
    *  @return Reference to the result vector.
    */
  DVecType& scale(size_type inx, double x, size_type N=0);

  /**  Calculate the vector inner product between two substrings.
    *  @brief Vector inner product
    *  @param inx  Index of the first element in this vector to be multiplied.
    *  @param v    Vector to be multiplied
    *  @param inx2 index of first word in the current vector
    *  @param N    Number of words to be multiplied.
    *  @return Reference to the result vector.
    */
  double dot(size_type inx, const DVector& v, size_type inx2=0,
	     size_type N=0) const;

  /**  Calculate the complex inner product between two vector substrings.
    *  @brief Divide two vectors
    *  @param inx  Index of the first element in this vector to be multiplied.
    *  @param v    Vector to be multiplied
    *  @param inx2 index of first word in the current vector
    *  @param N    Number of words to be multiplied.
    *  @return Reference to the result vector.
    */
  dComplex cdot(size_type inx, const DVector& v, size_type inx2=0, 
		size_type N=0) const;

  /**  Divide <tt>N</tt> values starting at index <tt>inx</tt> of the 
    *  current vector by the values in the divisor vector starting at 
    *  index <tt>inx2</tt>.
    *  @brief Divide two vectors
    *  @param inx  Index of the first element in this vector to be multiplied.
    *  @param v    Vector to be multiplied
    *  @param inx2 index of first word in the current vector
    *  @param N    Number of words to be multiplied.
    *  @return Reference to the result vector.
    */
  DVecType& div(size_type inx, const DVector& v, size_type inx2=0, 
		size_type N=0);

  /**  The elements of a specified sub-vector are converted to complex 
    *  and summed.
    *  @brief Get the sum of a complex sub-vector.
    *  @return The sum of the specified elements.
    *  @param inx The first element to be summed.
    *  @param N   The number of elements to be summed.
    */
  dComplex CSum(size_type inx, size_type N)  const;

  /**  The elements of a specified sub-vector are converted to double 
    *  and summed.
    *  @brief Get the sum of a sub-vector.
    *  @return The sum of the specified elements.
    *  @param inx The first element to be summed.
    *  @param N   The number of elements to be summed.
    */
  double VSum(size_type inx, size_type N)  const;

  /**  Data from the rhs DVector are converted to the lhs DVector type
    *  and copied to the lhs Vector.
    *  @brief Assignment operator.
    */
  DVector& operator=(const DVector& rhs);

  /**  Data from the rhs DVector are converted to the lhs DVector type
    *  and compared to the lhs Vector.
    *  @brief Assignment operator.
    */
  bool operator==(const DVector& rhs) const;

  /**  A reference to the specified vector element is returned.
    *  @brief   Reference a Vector element.
    *  @return Reference a Vector element.
    *  @param  inx Index of element to be accessed.
    */
  T& operator[] (size_type inx);

  /**  A constant reference to the vector element is returned.
    *        Constant and non-constant references are implemented.
    *  @brief Reference a Vector element.
    *  @param inx Index of element to be accessed.
    */
  const T& operator[] (size_type inx) const;

  /**  The data in the DVType vector are formatted and copied to the 
    *        specified ostream.
    *  @brief Print the data.
    */
  std::ostream& Dump(std::ostream& out) const;

private:
  /**  Pointer to a data array of the appropriate type. The array is 
    *        allocated by and owned by the DVecType object.
    *  @brief Pointer to allocated vector
    */
  CWVec<T> mData;
};

//======================================  inline DVType method templates
#ifndef __CINT__

template<class T>
inline
DVecType<T>::DVecType(const CWVec<T>& cv, size_type i, size_type l)
  : mData(cv)
{
    mData.substr(i, l);
}

template<class T>
inline typename DVecType<T>::size_type 
DVecType<T>::getLength(void) const {
    return mData.size();
}

template<class T>
inline
DVecType<T>::DVecType(void)
{
}

//--------------------------------------  Non constant references.
template<class T>
inline T* 
DVecType<T>::refTData(void) { 
    return mData.ref();
}

template<class T>
inline const T* 
DVecType<T>::refTData(void) const { 
    return mData.ref();
}

template<class T>
inline bool
DVecType<T>::shared(void) const {
    return mData.shared();
}

//------------------------------------  operator[] templates.
template<class T>
inline T& 
DVecType<T>::operator[] (size_type inx) {
    return mData[inx];
}

template<class T>
inline const T& 
DVecType<T>::operator[] (size_type inx) const {
    return mData[inx];
}

template<class T>
inline typename DVecType<T>::size_type
DVecType<T>::capacity(void) const {
    return mData.capacity();
}

//======================================  Reserve a minimum data length.
template<class T>
inline void 
DVecType<T>::reserve(size_type len) {
    mData.reserve(len);
}

//======================================  Typed data retrieval templates
template<>
inline short
DVecType<fComplex>::getShort(size_type inx) const
{
    return short(mData[inx].Real());
}

template<>
inline short
DVecType<dComplex>::getShort(size_type inx) const
{
    return short(mData[inx].Real());
}

template<class T>
inline short
DVecType<T>::getShort(size_type inx) const
{
    return short(mData[inx]);
}

template<>
inline int
DVecType<fComplex>::getInt(size_type inx) const
{
    return int(mData[inx].Real());
}

template<>
inline int
DVecType<dComplex>::getInt(size_type inx) const
{
    return int(mData[inx].Real());
}

template<class T>
inline int
DVecType<T>::getInt(size_type inx) const
{
    return int(mData[inx]);
}

template<>
inline float
DVecType<fComplex>::getFloat(size_type inx) const
{
    return mData[inx].Real();
}

template<>
inline float
DVecType<dComplex>::getFloat(size_type inx) const
{
    return mData[inx].Real();
}

template<class T>
inline float
DVecType<T>::getFloat(size_type inx) const
{
    return float(mData[inx]);
}

template<>
inline double
DVecType<fComplex>::getDouble(size_type inx) const
{
    return mData[inx].Real();
}

template<>
inline double
DVecType<dComplex>::getDouble(size_type inx) const
{
    return mData[inx].Real();
}

template<class T>
inline double
DVecType<T>::getDouble(size_type inx) const
{
    return double(mData[inx]);
}

template<class T>
inline fComplex
DVecType<T>::getCplx(size_type inx) const
{
    return fComplex(mData[inx]);
}

template<class T>
inline dComplex
DVecType<T>::getDCplx(size_type inx) const
{
    return dComplex(mData[inx]);
}

template<class T>
inline DVector::DVType
DVecType<T>::getType(void) const {
    return getDataType();
}

typedef DVecType<short>    DVectS;
typedef DVecType<int>      DVectI;
typedef DVecType<float>    DVectF;
typedef DVecType<double>   DVectD;
typedef DVecType<fComplex> DVectC;
typedef DVecType<dComplex> DVectW;

#endif  // !def(__CINT__)

#endif  // !def(DVECTOR_HH)
