//
//    CorrInit produces binary  The Channel configuration 
//    for the strip chart is specified by an ascii file passed to Chart()
//    as an argument. The format of the file is as follows:
//
//    Line 1: # Comments (start with a #)
//    Line 2: # More comments
//    ...  
//    Line n: # Last comment
//    Line n+1: <descriptor>
//    Line n+2: <number of channels>
//    Line n+3: <sample time>
//    Line n+4: <signal channel> <frequency> 
//    Line n+5: <first environmental channel> <frequency>
//    ...
//    Line n+<number of channels>+3: <last environmental channel> <frequency>
//    Line n+<number of channels>+4: <clean>
//
//    The fields above have the following meanings:
//
//    <descriptor>   A character string describing the run, used just for
//                      naming intermediate directories and files and the
//                      root canvases
//    <number of channels>   Number of channels including the signal 
//    <sample time>  The period in seconds over which correlations are sought.
//    <signal/environmental channel>      The name of the signal/environmental 
//                      channel between which correlations are sought.
//    <frequency>    The sample rate of the channel
//    <clean>        Set to 1 if you want a best estimate of the `true' spectrum
//                      based on a knowledge of correlations gleaned, 0 otherwise.
//
//    Note: The current implementation of DEnvCorr cannot handle
//    repeated channels.  Development to address this is ongoing.
//
#include <time.h>
#include "DCorrInit.hh"

const int CINT_MAX_CHANS=10;

class DCorrInit : public DatEnv {
  public:
  DCorrInit(int argc, const char* argv[]);
  ~DCorrInit();
  void ProcessData(void);
  private:
  const char* config;
  fComplex *fsdata;
  float delta_f;
  float dTime;
  int nChan;
  int i;
  char filename[128];
  char descriptor[128];
  char fft_dir[128];
#ifndef __CINT__
  Chan *chanv;
#else 
  Chan chanv[CINT_MAX_CHANS];  
#endif 
};     

using namespace std;

void
CorrInit(const char* config) {
    DCorrInit xxx(1, &config);
    xxx.MainLoop();
}


#ifndef __CINT__
EXECDAT(DCorrInit);
#endif

DCorrInit::DCorrInit(int argc, const char* argv[]) 
  :DatEnv(argc, argv) {

        //---------------------------------  Check file name
  config = argv[argc-1];
  if (!config || !*config || argc!=2) {
    cerr<< "CorrInit: configuration file not specified" << endl;
    finish();
    return;
  }

   //---------------------------------  Get parameters.
   ifstream ifs(config, ios::in);
   if (ifs.bad()) {
     cerr<< "CorrInit: Unable to open configuration file " << config << endl;
	finish();
     return;
   }

   cout << "# CorrInit: Configuration file is " << config << endl;

   while (1) {
     ifs.getline(descriptor, sizeof(descriptor));
     if (descriptor[0] == '#') continue;   // Ignore any preceeding comments
     cerr << "# CorrInit: Descriptor is " <<  descriptor << endl;
	break;
   }
     
   ifs >> nChan;
   cerr << "# CorrInit: Number of channels  (including signal) = " << nChan << endl;


#ifndef __CINT__
  chanv=new Chan[nChan];
#else 
   if (nChan > CINT_MAX_CHANS) {
    cerr << "# CorrInit: CINT can only handle a maximum of CINT_MAX_CHANS = " 
	    << CINT_MAX_CHANS << " channels" << endl;
    cerr << "         Run outside root or increase CINT_MAX_CHANS at head of DCorrInit.cc" << endl;
     abort();
   } 
#endif

   ifs >> dTime;
   cerr << "# CorrInit: Sample time = " << dTime << endl;

   for (int id=0 ; id<nChan ; id++) {
     ifs >>  chanv[id].mName >> chanv[id].mRate;
	chanv[id].mPoint=int(ceil(dTime*chanv[id].mRate));
     chanv[id].mTS = new TSeries(Time(0), Interval(0.0), chanv[id].mPoint);
     chanv[id].mFS = new FSeries;
     mIn.addChannel(chanv[id].mName, 0, &(chanv[id].mTS));
     cout << "# CorrInit: Added channel " << chanv[id].mName << "\t No of points :" << chanv[id].mPoint << endl;  
   }
   ifs.close();

   mIn.setBuffer(0);    // Take up data as requested
   mIn.setDebug(0);
   mIn.setStride(dTime);
   cerr << "# CorrInit: Starting from file " << mIn.getFile() << endl;
}

 
void DCorrInit::ProcessData() {
  TSeries* ts;
  FSeries* fs;


  delta_f=chanv[0].mRate/chanv[0].mPoint;

  //  set up working directory based on name of descriptor
  sprintf(fft_dir,"%s_fft",descriptor);

  char cmd[128];
  sprintf(cmd,"mkdir -p %s",fft_dir); // -p suppresses errors if dir exists

  cerr << "# CorrInit: Executing system command:   " << cmd << endl;
  system(cmd);
  cerr << "# CorrInit: Placing files in directory " << fft_dir  << endl;

  int length=chanv[0].mPoint;
  for(int id=1;id<nChan;id++) {
    if (chanv[id].mPoint<length) length=chanv[id].mPoint;
  }
  length/=2;   // because we are creating separate arrays for the real and imaginary parts

  for (int id=0;id<nChan;id++) {
    // check frames are consistent with configuration file 
    if (chanv[id].mRate*double(chanv[id].mTS->getTStep()) != 1.0) {
	 cerr << "# CorrInit: Sample rate for " << chanv[id].mName <<  " is " <<
	   1.0/double(chanv[id].mTS->getTStep()) << " not " <<
	   chanv[id].mRate << " as specified in " << config << endl;
	 abort();
    }

    // assign memory for storing data in fComplex array
    fsdata = new fComplex[chanv[id].mPoint/2];
    if (!fsdata) {
	 cerr<< "# CorrInit: unable to allocate memory for fsdata" << endl;
	 abort();
    }
 
    ts = chanv[id].mTS;
    fs = chanv[id].mFS;
    fs->setData(*ts);   // FFT  to an FSeries

    i=fs->getData(chanv[id].mPoint/2,fsdata);
    if (fs->getNStep()!=i) { 
	 cerr << "# CorrInit: fs, nbin = " <<  fs->getNStep() << "   points returned  = " << i << endl;
    }

    // assign memory for storing data in float array with NR conventions

    float *rp_fft,*ip_fft;
    rp_fft = new float[length];
    ip_fft = new float[length];
    if (!rp_fft || !ip_fft) {
	 cerr<< "# CorrInit: unable to allocate memory for ffts" << endl;
	 finish();
    }

    float relnorm;
    relnorm=0.5*float(chanv[id].mPoint/2);  // relative normalisation between NR and DMT

    for (int i=0;i<length;i++) {
	 rp_fft[i]=relnorm*fsdata[i].Real();
	 ip_fft[i]=-relnorm*fsdata[i].Imag();
    }

    // Print the Fourier transform to file 
    sprintf(filename,"%s/%s_fft.b",fft_dir,chanv[id].mName);
    write_binary_fft(filename,length,rp_fft,ip_fft,delta_f);

    delete [] fsdata;
    delete [] rp_fft;
    delete [] ip_fft;
  }
     
  finish();  // we only wish to process one data segment
}

DCorrInit::~DCorrInit() {
    //---------------------------------  Clean up at the end
    for (int id=0 ; id<nChan ; id++) {
        mIn.rmChannel(chanv[id].mName);
	   delete chanv[id].mTS;
	   delete chanv[id].mFS;
    }

    cout << "CorrInit finished" << endl;
}


 
